@@ -40,18 +40,30 @@ LFSR_PRIMITIVE_POLYNOMIALS = \
                     16: [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1],
                     17: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                     18: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
-                    19: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1],
-                    20: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
-                    21: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
-                    22: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
-                    23: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
-                    24: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1],
-                    25: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
-                    26: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1],
-                    27: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1],
-                    28: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
-                    29: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
-                    30: [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1]
+                    19: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1,
+                         1],
+                    20: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
+                         0, 0],
+                    21: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+                         0, 1, 0],
+                    22: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+                         0, 0, 0, 1],
+                    23: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+                         1, 0, 0, 0, 0],
+                    24: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
+                         0, 0, 0, 0, 1, 1],
+                    25: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+                         0, 0, 0, 0, 1, 0, 0],
+                    26: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+                         0, 0, 1, 0, 0, 0, 1, 1],
+                    27: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+                         0, 0, 0, 0, 1, 0, 0, 1, 1],
+                    28: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+                         0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
+                    29: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+                         0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
+                    30: [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1]
                 }
             },
             3: {
@@ -101,8 +113,8 @@ class CVEPSpellerData(components.ExperimentData):
 
     def __init__(self, mode, paradigm_conf, commands_info, onsets, command_idx,
                  unit_idx, level_idx, matrix_idx, cycle_idx, trial_idx,
-                 cvep_model, spell_result, spell_target=None,
-                 **kwargs):
+                 cvep_model, spell_result, fps_resolution, spell_target=None,
+                 raster_events=None, **kwargs):
 
         # Check errors
         if mode not in ('train', 'test'):
@@ -121,7 +133,9 @@ class CVEPSpellerData(components.ExperimentData):
         self.trial_idx = trial_idx
         self.cvep_model = cvep_model
         self.spell_result = spell_result
+        self.fps_resolution = fps_resolution
         self.spell_target = spell_target
+        self.raster_events = raster_events
 
         # Optional attributes
         for key, value in kwargs.items():
@@ -369,39 +383,42 @@ class CVEPSpellerDataset(components.Dataset):
 # ---------------------------------- MODELS ---------------------------------- #
 class CVEPModelCircularShifting(components.Algorithm):
 
-    def __init__(self, refresh_rate=60, apply_filter_bank=False, art_rej=None,
+    def __init__(self, bpf=[[7, (1.0, 30.0)]], notch=[7, (49.0, 51.0)],
+                 art_rej=None, correct_raster_latencies=False,
                  *args, **kwargs):
         super().__init__()
-        # todo: settings?
-        if not apply_filter_bank:
-            # Preprocessing (bandpass and notch IIR filters)
-            self.add_method('prep_method',
-                            StandardPreprocessing(bpf_order=7,
-                                                  bpf_cutoff=(1.0, refresh_rate/2),
-                                                  notch_order=7,
-                                                  notch_cutoff=(49.0, 51.0)
-                                                  )
-                            )
+
+        if len(bpf) == 1:
+            if notch is not None:
+                self.add_method('prep_method', StandardPreprocessing(
+                    bpf_order=bpf[0][0], bpf_cutoff=bpf[0][1],
+                    notch_order=notch[0], notch_cutoff=notch[1]))
+            else:
+                self.add_method('prep_method', StandardPreprocessing(
+                    bpf_order=bpf[0][0], bpf_cutoff=bpf[0][1],
+                    notch_order=None, notch_cutoff=None))
         else:
-            # Filter bank
-            filter_bank = [{'order': 7,
-                            'cutoff': (1.0, 60.0),
-                            'btype': 'bandpass'},
-                           {'order': 7,
-                            'cutoff': (12.0, 60.0),
-                            'btype': 'bandpass'},
-                           {'order': 7,
-                            'cutoff': (30.0, 60.0),
-                            'btype': 'bandpass'},
-                           ]
-            self.add_method('prep_method',
-                            FilterBankPreprocessing(filter_bank=filter_bank,
-                                                    notch_order=7,
-                                                    notch_cutoff=(49.0, 51.0)
-                                                    )
-                            )
+            filter_bank = []
+            for i in range(len(bpf)):
+                filter_bank.append({
+                    'order': bpf[i][0],
+                    'cutoff': bpf[i][1],
+                    'btype': 'bandpass'
+                })
+            if notch is not None:
+                self.add_method('prep_method', FilterBankPreprocessing(
+                    filter_bank=filter_bank, notch_order=notch[0],
+                    notch_cutoff=notch[1]))
+            else:
+                self.add_method('prep_method', FilterBankPreprocessing(
+                    filter_bank=filter_bank, notch_order=None,
+                    notch_cutoff=None))
+
         # Feature extraction and classification (circular shifting)
-        self.add_method('clf_method', CircularShiftingClassifier(art_rej=art_rej))
+        self.add_method('clf_method', CircularShiftingClassifier(
+            art_rej=art_rej,
+            correct_raster_latencies=correct_raster_latencies
+        ))
 
     def check_predict_feasibility(self, dataset):
         return self.get_inst('clf_method')._is_predict_feasible(dataset)
@@ -470,12 +487,14 @@ class CVEPModelCircularShifting(components.Algorithm):
 
         # Extract the selected label using the maximum number of cycles
         selected_seq = 0  # todo: several sequences in the same matrix
-        spell_result = pred_item_by_no_cycles[-1][selected_seq]['label']
+        spell_result = pred_item_by_no_cycles[-1][selected_seq][
+            'sorted_cmds'][0]['label']
 
         # Extract the selected label depending on the number of cycles
         spell_result_per_cycle = {}
         for nc, pred in enumerate(pred_item_by_no_cycles):
-            spell_result_per_cycle[nc] = pred[selected_seq]['label']
+            spell_result_per_cycle[nc] = pred[selected_seq]['sorted_cmds'][
+                0]['label']
 
         # Create the decoding dictionary
         cmd_decoding = {
@@ -523,11 +542,12 @@ class StandardPreprocessing(components.ProcessingMethod):
                                             filt_method=self.filt_method)
         self.bpf_iir_filter.fit(fs)
         # Notch
-        self.notch_iir_filter = mds.IIRFilter(order=self.notch_order,
-                                              cutoff=self.notch_cutoff,
-                                              btype='bandstop',
-                                              filt_method=self.filt_method)
-        self.notch_iir_filter.fit(fs)
+        if self.notch_cutoff is not None:
+            self.notch_iir_filter = mds.IIRFilter(order=self.notch_order,
+                                                  cutoff=self.notch_cutoff,
+                                                  btype='bandstop',
+                                                  filt_method=self.filt_method)
+            self.notch_iir_filter.fit(fs)
 
     def transform_signal(self, signal):
         """Transforms an EEG signal applying IIR filterings
@@ -538,7 +558,8 @@ class StandardPreprocessing(components.ProcessingMethod):
             Signal to transform. Shape [n_samples x n_channels]
         """
         signal = self.bpf_iir_filter.transform(signal)
-        signal = self.notch_iir_filter.transform(signal)
+        if self.notch_iir_filter is not None:
+            signal = self.notch_iir_filter.transform(signal)
         return signal
 
     def fit_transform_signal(self, signal, fs):
@@ -666,11 +687,12 @@ class FilterBankPreprocessing(components.ProcessingMethod):
             self.filter_bank_iir_filters.append(iir)
 
         # Notch
-        self.notch_iir_filter = mds.IIRFilter(order=self.notch_order,
-                                              cutoff=self.notch_cutoff,
-                                              btype='bandstop',
-                                              filt_method=self.filt_method)
-        self.notch_iir_filter.fit(fs)
+        if self.notch_cutoff is not None:
+            self.notch_iir_filter = mds.IIRFilter(order=self.notch_order,
+                                                  cutoff=self.notch_cutoff,
+                                                  btype='bandstop',
+                                                  filt_method=self.filt_method)
+            self.notch_iir_filter.fit(fs)
 
     def transform_signal(self, signal):
         """Transforms an EEG signal applying the filter bank
@@ -680,10 +702,11 @@ class FilterBankPreprocessing(components.ProcessingMethod):
         signal: np.array or list
             Signal to transform. Shape [n_samples x n_channels]
         """
+        if self.notch_iir_filter is not None:
+            signal = self.notch_iir_filter.transform(signal)
         signals = []
         for filter in self.filter_bank_iir_filters:
             signal_ = signal.copy()
-            signal_ = self.notch_iir_filter.transform(signal_)
             signals.append(filter.transform(signal_))
         return signals
 
@@ -751,7 +774,7 @@ class CircularShiftingClassifier(components.ProcessingMethod):
     it computes a template for each sequence.
     """
 
-    def __init__(self, art_rej=None, **kwargs):
+    def __init__(self, correct_raster_latencies=False, art_rej=None, **kwargs):
         """ Class constructor """
         super().__init__(fit_dataset=['templates',
                                       'cca_by_seq'])
@@ -759,6 +782,7 @@ class CircularShiftingClassifier(components.ProcessingMethod):
         self.fitted = dict()
 
         self.art_rej = art_rej
+        self.correct_raster_latencies = correct_raster_latencies
 
     def _assert_consistency(self, dataset: CVEPSpellerDataset):
         len_seqs = set()
@@ -884,7 +908,8 @@ class CircularShiftingClassifier(components.ProcessingMethod):
                             epochs_by_seq[tuple(seq_)][filter_idx] = \
                                 epochs[ep_idxs_, :, :]
                         else:
-                            epochs_by_seq[tuple(seq_)][filter_idx] = np.concatenate((
+                            epochs_by_seq[tuple(seq_)][
+                                filter_idx] = np.concatenate((
                                 epochs_by_seq[tuple(seq_)][filter_idx],
                                 epochs[ep_idxs_, :, :]
                             ), axis=0)
@@ -919,12 +944,14 @@ class CircularShiftingClassifier(components.ProcessingMethod):
                     epoch_to_keep = np.zeros(epochs_std.shape)
                     for i in range(len(ch_std)):
                         epoch_to_keep[:, i] = (
-                                (epochs_std[:, i] < (np.median(epochs_std[:, i]) +
-                                                     std_epoch_rejection * ch_std[
-                                                         i])) &
-                                (epochs_std[:, i] > (np.median(epochs_std[:, i]) -
-                                                     std_epoch_rejection * ch_std[
-                                                         i]))
+                                (epochs_std[:, i] < (
+                                            np.median(epochs_std[:, i]) +
+                                            std_epoch_rejection * ch_std[
+                                                i])) &
+                                (epochs_std[:, i] > (
+                                            np.median(epochs_std[:, i]) -
+                                            std_epoch_rejection * ch_std[
+                                                i]))
                         )
                     # Keep only epochs that are suitable for all channels
                     idx_to_keep = (
@@ -940,7 +967,8 @@ class CircularShiftingClassifier(components.ProcessingMethod):
 
                 # Reference (main template repeated 'no_cycles' times)
                 main_template = np.mean(epochs_by_seq[seq_][filter_idx], axis=0)
-                R = np.tile(main_template.T, epochs_by_seq[seq_][filter_idx].shape[0]).T
+                R = np.tile(main_template.T,
+                            epochs_by_seq[seq_][filter_idx].shape[0]).T
 
                 # Input data (concatenated epochs)
                 X = epochs_by_seq[seq_][filter_idx]
@@ -955,7 +983,8 @@ class CircularShiftingClassifier(components.ProcessingMethod):
                 for lag in range(len(seq_)):
                     lag_samples = int(np.round(lag / fps_resolution * fs))
                     lagged_seq = np.roll(seq_, -lag, axis=0)
-                    lagged_template = np.roll(main_template, -lag_samples, axis=0)
+                    lagged_template = np.roll(main_template, -lag_samples,
+                                              axis=0)
                     templates[tuple(lagged_seq)] = lagged_template
 
                     if show_progress_bar:
@@ -1061,6 +1090,17 @@ class CircularShiftingClassifier(components.ProcessingMethod):
             idx = (np.array(exp_data.trial_idx) == trial_idx) & \
                   (np.array(exp_data.cycle_idx) <= nc)
 
+            # Raster latencies?
+            raster_dict = None
+            if self.correct_raster_latencies:
+                possible_onsets_idx = np.where(
+                    exp_data.raster_events['onset'] <
+                    exp_data.onsets[idx][-1]
+                )[0]
+                if possible_onsets_idx.size > 0:
+                    raster_dict = exp_data.raster_events['event'][
+                        possible_onsets_idx[-1]]
+
             # For each fitted sequence
             pred_item = []
             for seq_, seq_data_ in self.fitted['sequences'].items():
@@ -1088,7 +1128,8 @@ class CircularShiftingClassifier(components.ProcessingMethod):
                             np.std(epochs[:, :len_epoch_sam, :], axis=1)
                         for i in range(epoch_std_by_channel.shape[0]):
                             discard_epoch = epoch_std_by_channel[i, :] > \
-                                            seq_data_[filter_idx]['std_by_channel'] \
+                                            seq_data_[filter_idx][
+                                                'std_by_channel'] \
                                             * self.art_rej
                             if np.any(discard_epoch):
                                 # TODO: precompute distance matrix before
@@ -1096,7 +1137,8 @@ class CircularShiftingClassifier(components.ProcessingMethod):
                                     self._interpolate_epoch(
                                         epoch=epochs[i, :len_epoch_sam, :],
                                         channel_set=sig_data.channel_set,
-                                        bad_channels_idx=np.where(discard_epoch)[0],
+                                        bad_channels_idx=
+                                        np.where(discard_epoch)[0],
                                         no_neighbors=3
                                     )
 
@@ -1112,11 +1154,14 @@ class CircularShiftingClassifier(components.ProcessingMethod):
                     seqs = []
                     for shift_seq_, template_ in \
                             seq_data_[filter_idx]['templates'].items():
-                        temp_p = np.dot(template_, x_) / \
-                                 np.sqrt(np.dot(np.dot(template_, template_),
-                                                np.dot(x_, x_)
-                                                )
-                                         )
+                        # Correct template using raster latencies
+                        lat_s = 0
+                        if raster_dict is not None:
+                            if shift_seq_ in raster_dict:
+                                lat_s = int(raster_dict[shift_seq_] * fs)
+                        tem_ = np.roll(template_, lat_s)
+                        temp_p = np.dot(tem_, x_) / np.sqrt(np.dot(np.dot(
+                            tem_, tem_), np.dot(x_, x_)))
                         corrs.append(temp_p)
                         seqs.append(shift_seq_)
                     f_corrs.append(corrs)
@@ -1131,17 +1176,14 @@ class CircularShiftingClassifier(components.ProcessingMethod):
                 sorted_seqs = seqs[sorted_idx, :]
 
                 # Identify the selected command
-                sel_cmd, sel_coords, sel_p = get_item_by_selected_sequence(
+                sorted_cmds = get_items_by_sorted_sequences(
                     experiment=exp_data,
                     trial_idx=trial_idx,
                     sorted_seqs=sorted_seqs,
                     sorted_corrs=sorted_corrs
                 )
                 pred_item.append({
-                    'coords': sel_coords,
-                    'label': sel_cmd['label'],
-                    'correlation': sel_p,
-                    'item': sel_cmd,
+                    'sorted_cmds': sorted_cmds,
                     'fitted_sequence': seq_
                 })
 
@@ -1238,13 +1280,13 @@ def check_if_shifted(seq1, seq2):
         return False
 
 
-def get_item_by_selected_sequence(experiment: CVEPSpellerData,
+def get_items_by_sorted_sequences(experiment: CVEPSpellerData,
                                   trial_idx, sorted_seqs, sorted_corrs=None):
     # Find the first index of the trial to access matrix_idx, etc
     try:
         idx = list(experiment.trial_idx == trial_idx).index(True)
     except ValueError as e:
-        raise ValueError('[get_item_by_selected_sequence] Trial with idx %i not'
+        raise ValueError('[get_items_by_sorted_sequences] Trial with idx %i not'
                          ' found in the experiment data! ' + str(e) % trial_idx)
 
     # Get the possible commands
@@ -1254,32 +1296,33 @@ def get_item_by_selected_sequence(experiment: CVEPSpellerData,
     possible_cmd = experiment.paradigm_conf[m_][l_][u_]
 
     # For each sequence in descending order of probability of being selected
-    selected_command = None
-    selected_corr = None
-    selected_coords = None
+    sorted_commands = list()
     for i in range(sorted_seqs.shape[0]):
         # For each possible command
+        curr_comm_dict = dict()
         for cmd_id in possible_cmd:
             cmd_seq = experiment.commands_info[m_][str(cmd_id)]['sequence']
             if np.all(np.array(cmd_seq) == sorted_seqs[i, :]):
-                selected_command = experiment.commands_info[m_][str(cmd_id)]
-                selected_coords = [m_, l_, u_, int(cmd_id)]
+                # Found!
+                curr_comm_dict['item'] = experiment.commands_info[m_][str(
+                    cmd_id)]
+                curr_comm_dict['label'] = curr_comm_dict['item']['label']
+                curr_comm_dict['coords'] = [m_, l_, u_, int(cmd_id)]
+                curr_comm_dict['correlation'] = None
                 if sorted_corrs is not None:
-                    selected_corr = sorted_corrs[i]
+                    curr_comm_dict['correlation'] = sorted_corrs[i]
+                sorted_commands.append(curr_comm_dict)
                 break
-        else:
-            continue    # only executed if the inner loop did NOT break
-        break           # only executed if the inner loop DID break
-    return selected_command, selected_coords, selected_corr
+    return sorted_commands
 
 
 def autocorr_zeropad(x):
     """ With zero padding, equivalent to np.correlate() """
     N = len(x)
     rxx = []
-    x_lagged = np.concatenate((np.zeros((N-1,)), x, np.zeros((N,))) )
-    for i in range(2*N-1):
-        rxx.append(np.sum(x * x_lagged[i:i+N]))
+    x_lagged = np.concatenate((np.zeros((N - 1,)), x, np.zeros((N,))))
+    for i in range(2 * N - 1):
+        rxx.append(np.sum(x * x_lagged[i:i + N]))
     rxx = np.array(rxx)
     return rxx
 
@@ -1288,7 +1331,7 @@ def autocorr_circular(x):
     """ With circular shifts (periodic correlation) """
     N = len(x)
     rxx = []
-    for i in range(-(N-1), N):
+    for i in range(-(N - 1), N):
         rxx.append(np.sum(x * np.roll(x, i)))
     rxx = np.array(rxx)
     return rxx
@@ -1369,5 +1412,5 @@ class LFSR:
             if base == 2:
                 sequence = np.array(sequence) * 2 - 1
             else:
-                sequence = np.array(sequence) - np.floor(base/2).astype(int)
+                sequence = np.array(sequence) - np.floor(base / 2).astype(int)
         return sequence
